// $Id: Util.java 16502 2012-10-25 14:38:58Z mike $

package org.faceless.pdf2.viewer2;

import java.io.UnsupportedEncodingException;
import java.util.*;
import java.beans.*;
import java.net.*;
import java.io.*;
import javax.swing.*;
import java.awt.*;
import java.applet.Applet;
import java.awt.event.*;
import org.faceless.util.log.*;

/**
 * A utility class that handles localized Strings, display of
 * error messages and so on. Developers extending the viewer are
 * welcome to use this class
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @since 2.11 - created from the SuperJOptionPane class
 */
public final class Util
{
    public static final BFOLogger LOGGER = new BFOLogger("org.faceless.pdf2.viewer2");

    private Util() {
    }

    /**
     * Display an Error message, including the stack trace
     * @param throwable the Throwable object this error relates to
     * @param parent the parent component (not used)
     */
    public static void displayThrowable(final Throwable throwable, Component parent) {
        if (!SwingUtilities.isEventDispatchThread()) {
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    displayThrowable(throwable, null);
                }
            });
            return;
        }
        Util.initialize();
        String title = UIManager.getString("PDFViewer.Error");

        final JPanel panel = new JPanel(new BorderLayout());
        final JLabel label = new JLabel(getMessage(throwable, false));
        panel.add(label);
        final JButton detailbutton = new JButton(UIManager.getString("PDFViewer.Show"));
        JOptionPane pane = new JOptionPane(panel, JOptionPane.ERROR_MESSAGE, JOptionPane.YES_NO_OPTION, null, new Object[] { detailbutton, UIManager.getString("PDFViewer.Close") });

        final JDialog dialog = pane.createDialog(null, title);
        dialog.setResizable(true);
        dialog.setModal(true);

        JEditorPane detail = new JEditorPane();
        detail.setContentType("text/html");
        detail.setText(getMessage(throwable, true));
        final JScrollPane scrollpane = new JScrollPane(detail);
        scrollpane.setPreferredSize(new Dimension(400, 200));

        detailbutton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                try {
                    if (detailbutton.getText().equals(UIManager.getString("PDFViewer.Show"))) {
                        panel.add(label, BorderLayout.NORTH);
                        panel.add(scrollpane);
                        detailbutton.setText(UIManager.getString("PDFViewer.Hide"));
                    } else {
                        panel.remove(scrollpane);
                        panel.add(label);
                        detailbutton.setText(UIManager.getString("PDFViewer.Show"));
                    }
                    dialog.pack();
                } catch (Throwable e) {
                    e.printStackTrace();
                }
            }
        });
        LOGGER.warning("PV1", "Viewer Exception", throwable);
        dialog.setVisible(true);
    }

    private static String getMessage(Throwable e, boolean expanded) {
        StringBuffer msg = new StringBuffer();
        if (expanded) {
            msg.append("<html><body>");
            while (e != null) {
                msg.append("<b>"+e.getClass().getName()+"</b>");
                if (e.getMessage()!=null) {
                    msg.append(": "+e.getMessage());
                }
                msg.append("<pre style='font-size:90%'>\n");
                StackTraceElement[] s = e.getStackTrace();
                for (int i=0;i<s.length;i++) {
                    msg.append("\u2022 " + s[i].getClassName() + ".");
                    msg.append("<b>"+s[i].getMethodName()+"</b>()");
                    if (s[i].getFileName() != null) {
                        msg.append(" at <tt>"+s[i].getFileName()+":"+s[i].getLineNumber()+"</tt>");
                    }
                    msg.append("\n");
                }
                msg.append("</pre>");
                e = e.getCause();
                if (e != null) {
                    msg.append("<br/><i>Caused by: </i>");
                }
            }
            msg.append("</body></html>");
        } else {
            String name = e.getClass().getName();
            name = name.substring(name.lastIndexOf('.') + 1);
            msg.append(name);
            if (e.getMessage()!=null) {
                msg.append(": "+e.getMessage());
            }
        }
        return msg.toString();
    }

    // To identify the Look & Feel - needed to work around bugs

    /**
     * Return true if the LAF uses the Aqua toolkit (OS X)
     */
    public static boolean isLAFAqua() {
        return UIManager.getLookAndFeel()!=null && (UIManager.getLookAndFeel().getClass().getName().equals("apple.laf.AquaLookAndFeel") || UIManager.getLookAndFeel().getClass().getName().equals("com.apple.laf.AquaLookAndFeel"));
    }

    /**
     * Return true if the LAF uses the GTK+ toolkit (UNIX)
     */
    public static boolean isLAFGTK() {
        return UIManager.getLookAndFeel()!=null && UIManager.getLookAndFeel().getClass().getName().equals("com.sun.java.swing.plaf.gtk.GTKLookAndFeel");
    }

    /**
     * Return true if the LAF uses the Nimbus toolkit
     */
    public static boolean isLAFNimbus() {
        String s;
        return UIManager.getLookAndFeel()!=null && ((s=UIManager.getLookAndFeel().getClass().getName()).equals("com.sun.java.swing.plaf.nimbus.NimbusLookAndFeel") || s.equals("javax.swing.plaf.nimbus.NimbusLookAndFeel"));
    }

    /**
     * Return true if the LAF uses the Window toolkit
     */
    public static boolean isLAFWindows() {
        return UIManager.getLookAndFeel()!=null && UIManager.getLookAndFeel().getClass().getName().equals("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");
    }

    /**
     * Return true if the LAF uses the Metal toolkit
     */
    public static boolean isLAFMetal() {
        return UIManager.getLookAndFeel()!=null && UIManager.getLookAndFeel().getClass().getName().equals("javax.swing.plaf.metal.MetalLookAndFeel");
    }

    /**
     * Initialize whatever needs initializing before any Swing objects are created.
     * Currently this just loads the locqlized Strings into the UIManager.
     */
    synchronized static void initialize() {
        if (UIManager.get("PDFViewer.NextPage")==null) {
            // To work around UIDefaults.addResourceBundle() not working in applets or JWS (newly filed bug)
            ResourceBundle bundle = ResourceBundle.getBundle("org.faceless.pdf2.viewer2.resources.Text");
            for (Enumeration e = bundle.getKeys();e.hasMoreElements();) {
                String key = (String)e.nextElement();
                if (UIManager.get(key)==null) {
                    UIManager.put(key, bundle.getString(key));
                }
            }
        }
    }

    /**
     * Return true if the viewer is running as an Applet inside a web browser, false otherwise
     * @param c the Component
     * @since 2.11.1
     */
    public static boolean isBrowserApplet(Component c) {
        Applet applet = (Applet)SwingUtilities.getAncestorOfClass(Applet.class, c);
        return applet!=null && !(applet instanceof PDFViewerApplet && ((PDFViewerApplet)applet).isDraggedFromBrowser());
    }

    /**
     * Return true if the viewer is running as an Applet that has been dragged outside of the
     * web browser, false otherwise.
     * @param c the Component
     * @since 2.11.1
     */
    public static boolean isStandaloneApplet(Component c) {
        PDFViewerApplet applet = (PDFViewerApplet)SwingUtilities.getAncestorOfClass(PDFViewerApplet.class, c);
        return applet!=null && applet.isDraggedFromBrowser();
    }

    /**
     * Return true if the specified component is running in a JNLP environment, false otherwise
     * @param c the Component
     * @since 2.11.1
     */
    public static boolean isJNLP(Component c) {
        try {
            return javax.jnlp.ServiceManager.lookup("javax.jnlp.BasicService")!=null;
        } catch (Throwable e) {
            return false;
        }
    }

    /**
     * Display a URL in a web browser, if possible
     * @param url the URL to open
     * @param c the Component
     * @return true if the URL could be opened, false otherwise
     * @since 2.11.1
     */
    public static boolean openURL(URL url, Component c) {
        String msg = UIManager.getString("PDFViewer.ConfirmOpen").replaceAll("\\{1\\}", url.toString());
        if (JOptionPane.showConfirmDialog(null, msg, msg, JOptionPane.YES_NO_OPTION)==JOptionPane.NO_OPTION) {
            return false;
        }
        try {
            java.awt.Desktop.getDesktop().browse(url.toURI());                                  // Java16
            return true;
        } catch (Throwable e) {
            try {
                Applet applet = (Applet)SwingUtilities.getAncestorOfClass(Applet.class, c);     // Applet
                if (applet!=null) {
                    applet.getAppletContext().showDocument(url);
                    return true;
                }
                try {                                                                           // JNLP
                    Object svc = Class.forName("javax.jnlp.ServiceManager").getMethod("lookup", new Class[] { String.class }).invoke(null, new Object[] { "javax.jnlp.BasicService" });
                    svc.getClass().getMethod("showDocument", new Class[] { URL.class }).invoke(svc, new Object[] { url });
                    return true;
                } catch (Throwable e2) {}
                String osname = System.getProperty("os.name");                                  // Application
                if (osname.startsWith("Mac OS")) {
                    Class.forName("com.apple.eio.FileManager").getDeclaredMethod("openURL", new Class[] { String.class }).invoke(null, new Object[] { url.toString() });
                    return true;
                } else if (osname.startsWith("Windows")) {
                    Runtime.getRuntime().exec("rundll32 url.dll,FileProtocolHandler " + url);
                    return true;
                }
            } catch (Throwable e2) {}
        }
        return false;
    }

    /**
     * Replace the fields {1}, {2} etc. in a String with their respective fields
     * @param message the message, eg "Loading {1} from {2}"
     * @param fields an array of N number of fields
     * @return the modified string
     */
    private static String replaceString(String message, String[] fields) {
        int i = 1, j;
        boolean found = false;
        while (i<=fields.length) {
            while ((j=message.indexOf("{"+i+"}"))>=0) {
                message = message.substring(0, j) + fields[i-1] + message.substring(j+3);
                found = true;
            }
            if (found) {
                found = false;
                i++;
            } else {
                break;
            }
        }
        return message;
    }

    /**
     * Return a UIManager string field with the specified fields replaced
     * @param key the message key,
     * @param field1 the field to be substituted for {1}
     * @return the modified string
     */
    public static String getUIString(String key, String field1) {
        return replaceString(UIManager.getString(key), new String[] { field1 });
    }

    /**
     * Return a UIManager string field with the specified fields replaced
     * @param key the message key,
     * @param field1 the field to be substituted for {1}
     * @param field2 the field to be substituted for {2}
     * @return the modified string
     */
    public static String getUIString(String key, String field1, String field2) {
        return replaceString(UIManager.getString(key), new String[] { field1, field2 });
    }

    /**
     * Remove the pageup/pagedown key bindings from ScrollPanes - they're clashing
     * with the pageup/pagedown we're mapping on the DocumentPanel
     */
    public static void fixScrollPaneKeyBindings(JScrollPane scrollpane) {
        InputMap inputmap;
        /*
        inputmap = scrollpane.getInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT).getParent();
        inputmap.remove(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_UP, 0));
        inputmap.remove(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_DOWN, 0));
        inputmap.remove(KeyStroke.getKeyStroke(KeyEvent.VK_HOME, 0));
        inputmap.remove(KeyStroke.getKeyStroke(KeyEvent.VK_END, 0));
        */
        inputmap = scrollpane.getInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_UP, 0), "dummyAction");
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_DOWN, 0), "dummyAction");
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_HOME, 0), "dummyAction");
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_END, 0), "dummyAction");
    }

    /**
     * Attempt to convert a relative URL to an absolute one.
     * URL is assumed to be relative to the file. If
     * we don't have that info but we're loaded from
     * an applet, then attempt to load URL relative to
     * the applet base. Otherwise, take your best shot.
     */
    public static URL toURL(DocumentPanel docpanel, String path) {
        try {
            URL url;
            File base = (File)docpanel.getClientProperty("file");
            if (base != null) {
                url = base.getParentFile().toURI().resolve(path).toURL();
            } else {
                Applet applet = (Applet)SwingUtilities.getAncestorOfClass(Applet.class, docpanel);
                if (applet != null) {
                    url = new URL(applet.getDocumentBase(), path);
                } else {
                    url = new URL(path);
                    if (url.getProtocol() == null) {
                        url = new File(path).toURI().toURL();
                    }
                }
            }
            return url;
        } catch (MalformedURLException e) {
            throw new IllegalArgumentException("Invalid URL \""+path+"\"", e);
        }
    }

    /**
     * Return true if the viewer can access the local filesystem
     */
    public static boolean hasFilePermission() {
        try {
            new JFileChooser(new File(System.getProperty("user.home")));
            return true;
        } catch (Throwable e) {
            return false;
        }
    }

    /**
     * Return the standard #RRGGBB representation of a Color
     */
    public static String encodeColor(Color c) {
        char[] s = "0123456789ABCDEF".toCharArray();
        return "#"+s[c.getRed()>>4]+s[c.getRed()&15]+s[c.getGreen()>>4]+s[c.getGreen()&15]+s[c.getBlue()>>4]+s[c.getBlue()&15];
    }

    /**
     * Get the screen resolution. Same as Toolkit.getScreenResolution, except
     * it allows us to override it using the PDFViewer "dpi" property
     */
    public static int getScreenResolution(Component comp) {
        int dpi = 0;
        if (comp != null && !(comp instanceof PDFViewer)) {
            comp = SwingUtilities.getAncestorOfClass(PDFViewer.class, comp);
        }
        if (comp != null) {
            String prop = ((PDFViewer)comp).getPropertyManager().getProperty("dpi");
            if (prop != null) {
                try {
                    dpi = Integer.parseInt(prop);
                } catch (Exception e) { }
            }
        }
        if (dpi == 0) {
            dpi = comp == null ? Toolkit.getDefaultToolkit().getScreenResolution() : comp.getToolkit().getScreenResolution();
        }
        return dpi;
    }

}
