// $Id: AnnotationLineFactory.java 17105 2013-03-15 18:12:19Z mike $

package org.faceless.pdf2.viewer2.feature;

import org.faceless.pdf2.viewer2.*;
import org.faceless.pdf2.*;
import org.faceless.pdf2.viewer2.util.ColorChoicePanel;
import javax.swing.*;
import java.awt.geom.*;
import javax.swing.event.*;
import javax.swing.undo.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.*;
import java.util.*;
import java.util.prefs.Preferences;

/**
 * Create annotations that handle Line {@link AnnotationShape} objects.
 *
 * <span class="featurename">The name of this feature is <span class="featureactualname">AnnotationLine</span></span>
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @since 2.11.7
 */
public class AnnotationLineFactory extends AnnotationShapeFactory {

    public AnnotationLineFactory() {
        super("AnnotationLine");
    }

    public String getAnnotationType() {
        return UIManager.getString("PDFViewer.annot.Line");
    }

    protected boolean matchesShape(AnnotationShape annot) {
        Shape shape = annot.getShape();
        return shape instanceof Line2D || shape instanceof GeneralPath;
    }

    public PDFAnnotation createNewAnnotation(float x1, float y1, float x2, float y2) {
        Line2D line = new Line2D.Float(x1, y1, x2, y2);
        AnnotationShape annot = new AnnotationShape(line);
        if (getViewer() != null) {
            annot.setAuthor(getViewer().getCurrentUser());
        }
        PDFStyle style = new PDFStyle();

        Color linecolor = Color.black;
        Color fillcolor = null;
        float weight = 1;
        float[] dash = null;
        String arrow1 = null;
        String arrow2 = null;
        Preferences preferences = getPreferences();
        if (preferences!=null) {
            linecolor = ColorChoicePanel.loadColor(preferences, "feature.AnnotationLineFactory.lineColor", linecolor);
            fillcolor = ColorChoicePanel.loadColor(preferences, "feature.AnnotationLineFactory.arrowColor", fillcolor);
            weight = preferences.getFloat("feature.AnnotationLineFactory.lineWeight", 1);
            dash = BorderStyleEditor.getDashArray(preferences.get("feature.AnnotationLineFactory.dashPattern", "solid"));
            arrow1 = preferences.get("feature.AnnotationLineFactory.arrow1", "None");
            arrow2 = preferences.get("feature.AnnotationLineFactory.arrow2", "None");
        }
        style.setLineColor(linecolor);
        style.setFillColor(fillcolor);
        style.setLineWeighting(weight);
        style.setLineDash(dash, 0);
        annot.setStyle(style);
        annot.setFirstLineEnding(arrow1);
        annot.setLastLineEnding(arrow2);
        return annot;
    }

    protected JComponent createEditor(final AnnotationShape annot, PDFParser parser) {
        PDFStyle style = annot.getStyle();
        JPanel editorPane = new JPanel();
        editorPane.setLayout(new GridBagLayout());

        addStockDetailsToEditComponent(annot, editorPane);

        GridBagConstraints gbc = new GridBagConstraints();
        gbc.weightx = 1;
        gbc.weighty = 1;
        gbc.fill = GridBagConstraints.BOTH;
        gbc.gridwidth = GridBagConstraints.REMAINDER;

        final JTextArea textarea = new JTextArea(5, 20);
        textarea.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent e) { }
            public void focusLost(FocusEvent e) {
                annot.setContents(textarea.getText());
            }
        });

        textarea.setEditable(true);
        textarea.setLineWrap(true);
        textarea.setText(annot.getContents());
        editorPane.add(new JScrollPane(textarea), gbc);

        gbc.weighty = 0;
        gbc.fill = gbc.NONE;
        JPanel appearancePanel = new JPanel(new FlowLayout(FlowLayout.CENTER, 0, 0));
        editorPane.add(appearancePanel, gbc);

        LineEndingSelector first = new LineEndingSelector(annot.getFirstLineEnding(), false) {
            public void itemStateChanged(ItemEvent e) {
                String end = getLineEnding();
                annot.setFirstLineEnding(end);
                super.itemStateChanged(e);
            }
        };
        first.setPreferences(getPreferences(), "feature.AnnotationLineFactory.arrow1");
        appearancePanel.add(first);

        BorderStyleEditor borderStyle = new BorderStyleEditor(annot, false);
        borderStyle.setPreferences(getPreferences(), "feature.AnnotationLineFactory.lineWeight", "feature.AnnotationLineFactory.dashPattern");
        appearancePanel.add(borderStyle);

        LineEndingSelector last = new LineEndingSelector(annot.getLastLineEnding(), true) {
            public void itemStateChanged(ItemEvent e) {
                String end = getLineEnding();
                annot.setLastLineEnding(end);
                super.itemStateChanged(e);
            }
        };
        last.setPreferences(getPreferences(), "feature.AnnotationLineFactory.arrow2");
        appearancePanel.add(last);

        gbc.weightx = 0;
        gbc.fill = GridBagConstraints.NONE;
        Color color = getLineColor(style);
        JComponent colorEditor = ColorChoicePanel.createColorChoiceButton(color, new ColorChoicePanel.ColorChoiceListener() {
            public void colorChosen(Color color) {
                PDFStyle style = annot.getStyle();
                style.setLineColor(color);
                annot.setStyle(style);
                Preferences preferences = getPreferences();
                if (preferences!=null) {
                    ColorChoicePanel.saveColor(preferences, "feature.AnnotationLineFactory.lineColor", color);
                }
            }
        }, ColorChoicePanel.TYPE_RECTANGLE, true, false, UIManager.getString("PDFViewer.annot.setColor"));
        appearancePanel.add(colorEditor);

        JPanel space = new JPanel();
        space.setPreferredSize(new Dimension(2, 20));
        appearancePanel.add(space);

        Color fillColor = getFillColor(style);
        JComponent fillColorEditor = ColorChoicePanel.createColorChoiceButton(fillColor, new ColorChoicePanel.ColorChoiceListener() {
            public void colorChosen(Color color) {
                PDFStyle style = annot.getStyle();
                style.setFillColor(color);
                annot.setStyle(style);
                Preferences preferences = getPreferences();
                if (preferences!=null) {
                    ColorChoicePanel.saveColor(preferences, "feature.AnnotationLineFactory.arrowColor", color);
                }
            }
        }, ColorChoicePanel.TYPE_ARROW, true, true, UIManager.getString("PDFViewer.annot.setColor"));
        appearancePanel.add(fillColorEditor);

        return editorPane;
    }

    public JComponent createComponent(final PagePanel pagepanel, PDFAnnotation annot) {
        final AnnotationShape shapeannot = (AnnotationShape)annot;

        ArrayList l = new ArrayList(4);
        float[] coords = new float[6];
        boolean tclosed = false;
        final BitSet ismove = new BitSet();
        for (PathIterator i = new FlatteningPathIterator(shapeannot.getShape().getPathIterator(null), 1);!i.isDone();i.next()) {
            int st;
            if ((st=i.currentSegment(coords))==i.SEG_CLOSE) {
                tclosed = true;
            } else {
                if (st == i.SEG_MOVETO) {
                    ismove.set(l.size());
                }
                l.add(new Point2D.Float(coords[0], coords[1]));
            }
        }
        final Point2D[] points = (Point2D[])l.toArray(new Point2D[0]);
        final boolean closed = tclosed;

        return new LineComponent(pagepanel, points, new boolean[points.length], closed) {
            public void showPopup(Point point) {
                popupPropertyMenu(shapeannot, this, point);
            }

            public void redrawAnnotation(boolean changed) {
                if (changed) {

                    Shape workpath;
                    if (points.length == 2) {
                        workpath = new Line2D.Float(points[0], points[1]);
                    } else {
                        GeneralPath path = new GeneralPath();
                        for (int i=0;i<points.length;i++) {
                            if (ismove.get(i)) {
                                path.moveTo((float)points[i].getX(), (float)points[i].getY());
                            } else {
                                path.lineTo((float)points[i].getX(), (float)points[i].getY());
                            }
                        }
                        if (closed) {
                            path.closePath();
                        }
                        workpath = path;
                    }

                    final Shape newpath = workpath;
                    final Shape oldpath = shapeannot.getShape();
                    shapeannot.setShape(newpath);
                    shapeannot.rebuild();

                    final DocumentPanel docpanel = pagepanel.getDocumentPanel();
                    docpanel.fireUndoableEditEvent(new UndoableEditEvent(docpanel, new AbstractUndoableEdit() {
                        public String getPresentationName() {
                            return UIManager.getString("InternalFrameTitlePane.moveButtonText");
                        }
                        public void undo() {
                            super.undo();
                            shapeannot.setShape(oldpath);
                            shapeannot.rebuild();
                        }
                        public void redo() {
                            super.redo();
                            shapeannot.setShape(newpath);
                            shapeannot.rebuild();
                        }
                    }));
                }
            }
        };
    }
}
