// $Id: RotateClockwise.java 17146 2013-03-20 14:47:37Z mike $

package org.faceless.pdf2.viewer2.feature;

import org.faceless.pdf2.viewer2.*;
import org.faceless.pdf2.*;
import java.awt.*;
import java.util.*;
import java.util.List;
import javax.swing.*;
import javax.swing.undo.*;
import javax.swing.event.*;
import java.awt.event.*;

/**
 * Create a button that will rotate the page 90 degrees clockwise.
 *
 * <span class="featurename">The name of this feature is <span class="featureactualname">RotateClockwise</span></span>
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @since 2.8.3
 */
public class RotateClockwise extends ViewerWidget implements DocumentPanelListener, ThumbnailPanel.ThumbnailSelectionAction {

    private final int diff;

    public RotateClockwise() {
        this("RotateClockwise", 90);
    }

    protected RotateClockwise(String name, int diff) {
        super(name);
        setButton("Edit", "resources/icons/"+name+".png", "PDFViewer.tt."+name);
        setMenu("Edit\t"+name);
        this.diff = diff;
    }

    public void initialize(PDFViewer viewer) {
        super.initialize(viewer);
        viewer.addDocumentPanelListener(this);
        getComponent().setEnabled(false);
    }

    private boolean isActionEnabled(DocumentPanel docpanel) {
        // This is verging on farce - should be done with an Action as the primary.
        if (docpanel == null) {
            PDFViewer viewer = getViewer();
            docpanel = viewer == null ? null : viewer.getActiveDocumentPanel();
        }
        return docpanel != null && docpanel.hasPermission("Assemble");
    }

    public void documentUpdated(DocumentPanelEvent event) {
        String type = event.getType();
        DocumentPanel docpanel = event.getDocumentPanel();
        PDFViewer viewer = docpanel.getViewer();
        if ("activated".equals(type) || "permissionChanged".equals(type)) {
            boolean enabled = isActionEnabled(docpanel);
            Component comp = viewer.getNamedComponent("Button"+getName());
            if (comp != null) {
                comp.setEnabled(enabled);
            }
            comp = viewer.getNamedComponent("Menu"+getName());
            if (comp != null) {
                comp.setEnabled(enabled);
            }
        } else if ("deactivated".equals(type)) {
            Component comp = viewer.getNamedComponent("Button"+getName());
            if (comp != null) {
                comp.setEnabled(false);
            }
            comp = viewer.getNamedComponent("Menu"+getName());
            if (comp != null) {
                comp.setEnabled(false);
            }
        }
    }

    public void action(ViewerEvent event) {
        DocumentPanel dp = event.getDocumentPanel();
        SidePanel sp = dp.getSelectedSidePanel();
        if (sp instanceof ThumbnailPanel.ThumbnailSidePanel) {
            ThumbnailPanel.View view = ((ThumbnailPanel.ThumbnailSidePanel)sp).getView();
            Action action = getAction(view);
            if (action != null) {
                // Performs rotate on the pages selected in the thumbnail
                // panel
                action.actionPerformed(null);
                return;
            }
        }
        rotate(dp, Collections.singletonList(event.getDocumentPanel().getPage()), diff);
    }

    protected void rotate(final DocumentPanel docpanel, final Collection pages, final int diff) {
        for (Iterator i = pages.iterator();i.hasNext();) {
            PDFPage page = (PDFPage)i.next();
            page.setPageOrientation(page.getPageOrientation() + diff);
        }

        docpanel.fireUndoableEditEvent(new UndoableEditEvent(docpanel, new AbstractUndoableEdit() {
            public String getPresentationName() {
                return UIManager.getString("PDFViewer."+getName());
            }
            public void undo() {
                super.undo();
                undoredo(-diff);
            }
            public void redo() {
                super.redo();
                undoredo(diff);
            }
            private void undoredo(int diff) {
                for (Iterator i = pages.iterator();i.hasNext();) {
                    PDFPage page = (PDFPage)i.next();
                    page.setPageOrientation(page.getPageOrientation() + diff);
                }
            }
        }));
    }

    public boolean isButtonEnabledByDefault() {
        return false;
    }

    public Action getAction(final ThumbnailPanel.View view) {
        if (!view.isFactoryEditable()) {
            return null;
        }

        final AbstractAction action = new AbstractAction(UIManager.getString("PDFViewer."+getName())) {
            public void actionPerformed(ActionEvent event) {
                DocumentPanel docpanel = view.getDocumentPanel();
                List l = new ArrayList();
                final PDF pdf = docpanel.getPDF();
                for (int i=0;i<view.getComponentCount();i++) {
                    ThumbnailPanel.SinglePagePanel panel = (ThumbnailPanel.SinglePagePanel)view.getComponent(i);
                    if (panel.isSelected()) {
                        l.add(panel.getPage());
                    }
                }
                rotate(docpanel, l, diff);
            }
        };

        view.getDocumentPanel().addDocumentPanelListener(new DocumentPanelListener() {
            public void documentUpdated(DocumentPanelEvent event) {
                if ("permissionChanged".equals(event.getType())) {
                    action.setEnabled(isActionEnabled(event.getDocumentPanel()));
                }
            }
        });

        return action;
    }


}
